c    "Copyright 2010 Prof. Robert Dalrymple, Prof. M. Gomez Gesteira, Dr Benedict Rogers, 
c     Dr Alejandro Crespo, Dr Muthukumar Narayanaswamy, Dr Shan Zou, Dr Andrea Panizzo "
c
c    This file is part of SPHYSICS.
c
c    SPHYSICS is free software; you can redistribute it and/or modify
c    it under the terms of the GNU General Public License as published by
c    the Free Software Foundation; either version 3 of the License, or
c    (at your option) any later version.
c
c    SPHYSICS is distributed in the hope that it will be useful,
c    but WITHOUT ANY WARRANTY; without even the implied warranty of
c    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
c    GNU General Public License for more details.
c
c    You should have received a copy of the GNU General Public License
c    along with this program.  If not, see <http://www.gnu.org/licenses/>.

      subroutine celij(j1,j2,kind_p1,ini_kind_p2,lx2,ly2,lz2,id_call)
c
      include 'common.3D'       
       
      if(j2.lt.0)then
        print*
        print*,'ERROR in celij.f'
        print*,'j2.lt.0'
        print*,'j1 ',j1
        print*,'j2 ',j2
        print*,'id_call ',id_call
        print*,'rank_id ',rank_id
        stop
      endif
      
      do kind_p2=ini_kind_p2,2
        if(nc(j2,kind_p2).ne.0) then

        do ii=1,nc(j1,kind_p1)
          i = ibox(j1,kind_p1,ii)
         
          do jj=1,nc(j2,kind_p2)
           j = ibox(j2,kind_p2,jj)
           
           drx = xp(i) - xp(j)
           dry = yp(i) - yp(j)
           drz = zp(i) - zp(j)
           
           call periodicityCorrection
     &          (i,j,drx,dry,drz,lx2,ly2,lz2,id_call)

           rr2 = drx*drx + dry*dry + drz*drz
                        
           if(rr2.lt.rr2_min)then
             if(i_particleType(i).lt.4.and.i_particleType(j).lt.4)then
               goto 641
             end if
              print*
              print*,'ERROR In celij.f: rr2.lt.rr2_min'
              print*,'itime ',itime
              print*,'rank_id ',rank_id
              print*,'lx2,ly2,lz2,id_call ',lx2,ly2,lz2,id_call
              print*,'j1, j2 ',j1, j2
              print*,'kind_p1, kind_p2 ',kind_p1, kind_p2
              print*,'I, j ',I, j
              print*,'xp(I), yp(I), zp(I) ',xp(i), yp(i), zp(i)
              print*,'xp(j), yp(j), zp(j) ',xp(j), yp(j), zp(j)
              print*,'drx, dry, drz ',drx, dry, drz
              print*,'rr2     ',rr2
              print*,'rr2_min ',rr2_min
              print*,'fourh2 ',fourh2
              print*,'q=sqrt(rr2)/h =',sqrt(rr2)/h
              print*,'i_particleType(i) ',i_particleType(i)
              print*,'i_particleType(j) ',i_particleType(j)
              print*,'pm(I) ',pm(I)
              print*,'pm(j) ',pm(j)
              print*,'up(I), vp(I), wp(I) ',up(i), vp(i), wp(i)
              print*,'up(j), vp(j), wp(j) ',up(j), vp(j), wp(j)
              print*,'rhop(I), rdot(I) ',rhop(i), rdot(i)
              print*,'rhop(j), rdot(j) ',rhop(j), rdot(j)
              print*,'(ncx, ncy, ncz)_local ',
     &                 ncx_local, ncy_local, ncz_local
              print*,'xmin_rank, xmax_rank ',xmin_rank, xmax_rank
              print*,'ymin_rank, ymax_rank ',ymin_rank, ymax_rank
              print*,'zmin_rank, zmax_rank ',zmin_rank, zmax_rank
              call MPI_BARRIER(MPI_COMM_WORLD, ierr)   !Synchronise all processes to this point
              stop
           end if
           
           if(rr2.lt.fourh2.and.rr2.gt.1.e-18) then
             
             dux = up(i) - up(j)
             duy = vp(i) - vp(j)
             duz = wp(i) - wp(j)

c            Calculating kernel & Normalized Kernel Gradient
             call kernel(drx,dry,drz,i,j,j1,j2,rr2) 
             call kernel_correction(i,j)
             
c
c            ...  average density
c

             robar  = 0.5*( rhop(i) + rhop(j) )
             one_over_rhobar = 2.0/(rhop(i) + rhop(j))
c             
             cbar = 0.5*( cs(i)   + cs(j)   ) 
c
c            ...  inner product rv
c
             dot = drx*dux + dry*duy + drz*duz

c	      Used to calculate the time step due to viscosity

             visc_dt=max(dot/(rr2 + eta2),visc_dt) 
c
c            ...  pressure and viscous force (1992; 3.3)
c
c            pm(j) is mass of particle j
c
             p_v = pr(i) + pr(j)  !+  pi_visc



c           for tensile correction (Monaghan , JCP.  159 (2000) 290- 311)
c           Only to be activated with cubic spline kernel 
c
            if(index_tensile.eq.1) then

c   _____ Tensile correction
 
              fab=Wab*od_Wdeltap      !NOTE: We'll use a non-normalized
              fab=fab*fab             !kernel to calculate tensile correction
              fab=fab*fab             !It's the (Wab/Wdeltap)**4  of Monaghan's paper

              if (p(i).gt.0) then
                  Ra= 0.01 *pr(i)
              else
                  Ra= 0.2 *abs(pr(i))
              endif

              if (p(j).gt.0) then
                  Rb= 0.01 *pr(j)
              else
                  Rb= 0.2 *abs(pr(j))
              endif

              p_v = p_v+ (Ra+Rb)*fab

            endif

            if(i_particleType(i).eq.4.and.i_particleType(j).eq.4)then !FP-FP
               
              ax(i) = ax(i) - pm(j) * p_v * frxi
              ay(i) = ay(i) - pm(j) * p_v * fryi
              az(i) = az(i) - pm(j) * p_v * frzi

              ax(j) = ax(j) + pm(i) * p_v * frxj
              ay(j) = ay(j) + pm(i) * p_v * fryj
              az(j) = az(j) + pm(i) * p_v * frzj

!                 !- Debug code -
!                 i_particleCheck = 36599
!                 i_rankCheck = 6
!                 dpN = 0.05*vnorm_mass
! 599             format(a8,2(1x,i7),i3,3(1x,i2),f12.6)             
!                 !Place in separate ifs            
!                 if(rank_id.eq.i_rankCheck.and.i.eq.i_particleCheck)then
!                     idrx = nint(drx/dpN)
!                     idry = nint(dry/dpN)
!                     idrz = nint(drz/dpN)
!                     write(*,599)
!      &              'C:I,j ',i,j,id_call,idrx,idry,idrz,
!      &                        - pm(j) * p_v * frzi
! !                     if(j.eq.-122863)then
! !                       print*,'i, j ',i, j
! !                       print*,'xp,yp,zp(i) ',xp(i), yp(i), zp(i)
! !                       print*,'xp,yp,zp(j) ',xp(j), yp(j), zp(j)
! !                       print*,' drx ',drx
! !                       print*,' dry ',dry
! !                       print*,' drz ',drz
! !                       print*,'rr2 ',rr2
! !                       print*,'pm(j) ',pm(j)
! !                       print*,'p_v ',p_v
! !                       print*,'frxi ',frxi
! !                       print*,'rhop(i), p(i) ',rhop(i), p(i)
! !                       print*,'rhop(j), p(j) ',rhop(j), p(j)
! !                       print*,'pr(i) ',pr(i)
! !                       print*,'pr(j) ',pr(j)
! !                       print*,'index_tensile, (Ra+Rb)*fab ',
! !      &                        index_tensile, (Ra+Rb)*fab
! !                     endif
!                 endif 
!                 !Place in separate ifs            
!                 if(rank_id.eq.i_rankCheck.and.j.eq.i_particleCheck)then
!                     idrx = -nint((drx)/dpN)
!                     idry = -nint((dry)/dpN)
!                     idrz = -nint((drz)/dpN)
!                     write(*,599)
!      &              'C:i,J ',i,j,id_call,idrx,idry,idrz,
!      &                        + pm(i) * p_v * frzj
! 
!                     if(i.eq.21344)then
!                       print*,'i, j ',i, j
!                       print*,'xp,yp,zp(i) ',xp(i), yp(i), zp(i)
!                       print*,'xp,yp,zp(j) ',xp(j), yp(j), zp(j)
!                       print*,' drx ',drx
!                       print*,' dry ',dry
!                       print*,' drz ',drz
!                       print*,'rr2 ',rr2
!                       print*,'pm(i) ',pm(i)
!                       print*,'p_v ',p_v
!                       print*,'frzj ',frzj
!                       print*,'rhop(i), p(i) ',rhop(i), p(i)
!                       print*,'rhop(j), p(j) ',rhop(j), p(j)
!                       print*,'pr(i) ',pr(i)
!                       print*,'pr(j) ',pr(j)
!                       print*,'index_tensile, (Ra+Rb)*fab ',
!      &                        index_tensile, (Ra+Rb)*fab
!                     endif
!                 endif 
              call gradients_calc(i,j,dux,duy,duz)               

              call viscosity(dot,drx,dry,drz,dux,duy,duz,rr2,
     +               cbar,robar,one_over_rhobar,i,j,j1,j2,term2i,term2j)


c  ...  density acceleration 
c       
             dot2i = dux*frxi + duy*fryi + duz*frzi
             dot2j = dux*frxj + duy*fryj + duz*frzj
             ar(i) = ar(i) + pm(j)*dot2i
             ar(j) = ar(j) + pm(i)*dot2j

c  ...  Thermal Energy
c	       (Monaghan, JCP 110 (1994) 399- 406)
              
              
             term1i=0.5 * p_v *( frxi*dux+fryi*duy+frzi*duz)
             term1j=0.5 * p_v *( frxj*dux+fryj*duy+frzj*duz)

             aTE(i)=aTE(i)+pm(j) * (term1i+term2i)
             aTE(j)=aTE(j)+pm(i) * (term1j+term2j)

c
c  ...  XSPH correction (Monaghan 1994;  J. Comp. Phys. 110. Formula 2.6)
c
              pmj_Wab_over_rhobar = pm(j)*Wab*one_over_rhobar
              ux(i) = ux(i) - dux*pmj_Wab_over_rhobar  !pm(j) * dux * Wab / robar
              vx(i) = vx(i) - duy*pmj_Wab_over_rhobar  !pm(j) * duy * Wab / robar
              wx(i) = wx(i) - duz*pmj_Wab_over_rhobar  !pm(j) * duz * Wab / robar

              pmi_Wab_over_rhobar = pm(i)*Wab*one_over_rhobar
              ux(j) = ux(j) + dux*pmi_Wab_over_rhobar   !pm(i) * dux * Wab / robar
              vx(j) = vx(j) + duy*pmi_Wab_over_rhobar   !pm(i) * duy * Wab / robar
              wx(j) = wx(j) + duz*pmi_Wab_over_rhobar   !pm(i) * duz * Wab / robar

c  ...  Vorticity calculation

              if(ipoute.eq.1.and.i_vort.eq.1.and.
     &          i_particleType(i).gt.3.and.i_particleType(j).gt.3)then  !FP-FP
                  call vorticity_calc(i,j,dux,duy,duz)
              endif  
		   			
            else if(i_particleType(i).eq.4.and.
     &              i_particleType(j).lt.4)then  !FP-BP,   j is boundary particle
              call MonaghanBC(i,j,drx,dry,drz,dot,dux,duy,duz,
     &                        fxbMON,fybMON,fzbMON)   
              ax(i) = ax(i) + fxbMON
              ay(i) = ay(i) + fybMON
              az(i) = az(i) + fzbMON
              temp = pVol(i)*visc_wall*
     &           ((drx*frxi+dry*fryi+drz*frzi)/(rr2 + eta2))
              ax(i) = ax(i) + temp*dux
              ay(i) = ay(i) + temp*duy
              az(i) = az(i) + temp*duz

!                 !- Debug code -
!                 i_particleCheck = 6551   !15250
!                 irank_check = 1
!                 dpN = 0.01*vnorm_mass
! 599             format(a8,2(1x,i7),i3,3(1x,i2),4(1x,f10.6))             
!                 !Place in separate ifs            
!                 if(rank_id.eq.irank_check.and.i.eq.i_particleCheck)then
!                     idrx = nint(drx/dpN)
!                     idry = nint(dry/dpN)
!                     idrz = nint(drz/dpN)
!                     write(*,599)
!      &              'C:I,j ',i,j,id_call,idrx,idry,idrz,
!      &                        fxbMON,fybMON,fzbMON, temp
!                     if(j.eq.1410)then
!                       print*,'i ',i
!                       print*,'xp,yp,zp(i) ',xp(i), yp(i), zp(i)
!                       print*,'xp,yp,zp(j) ',xp(j), yp(j), zp(j)
!                       print*,' -drx ',-drx
!                       print*,' -dry ',-dry
!                       print*,' -drz ',-drz
!                       print*,'rr2 ',rr2
!                       print*,'pm(i) ',pm(i)
!                       print*,'p_v ',p_v
!                       print*,'frxj ',frxj
!                       print*,'rhop(i), p(i) ',rhop(i), p(i)
!                       print*,'rhop(j), p(j) ',rhop(j), p(j)
!                       print*,'pr(i) ',pr(i)
!                       print*,'pr(j) ',pr(j)
!                       print*,'index_tensile, (Ra+Rb)*fab ',
!      &                        index_tensile, (Ra+Rb)*fab
!                     endif
!                   endif 
            else if(i_particleType(j).eq.4.and.
     &              i_particleType(i).lt.4)then  !FP-BP,   i is boundary particle
              call MonaghanBC(j,i,-drx,-dry,-drz,dot,-dux,-duy,-duz,
     &                        fxbMON,fybMON,fzbMON)   
              ax(j) = ax(j) + fxbMON
              ay(j) = ay(j) + fybMON
              az(j) = az(j) + fzbMON
              temp = pVol(j)*visc_wall*
     &           ((drx*frxj+dry*fryj+drz*frzj)/(rr2 + eta2))
              ax(j) = ax(j) - temp*dux
              ay(j) = ay(j) - temp*duy
              az(j) = az(j) - temp*duz

!                 !- Debug code -
!                 i_particleCheck = 6551   !15250
!                 irank_check = 1
!                 dpN = 0.01*vnorm_mass
!                 !Place in separate ifs            
!                 if(rank_id.eq.irank_check.and.j.eq.i_particleCheck)then
!                   idrx = -nint((drx)/dpN)
!                   idry = -nint((dry)/dpN)
!                   idrz = -nint((drz)/dpN)
!                   write(*,599)
!      &            'C:i,J ',i,j,id_call,idrx,idry,idrz,
!      &                      fxbMON,fybMON,fzbMON, temp
! 
!                 endif 
            endif  ! Interaction fluid-fluid or fluid-boundary
		  
c           !-- Debug Code------
c           if(rank_id.eq.-10)then
c             i_particleCheck = 1831
c             if(i.eq.i_particleCheck)then
c                i_drx = nint(drx/(vnorm_mass*dpx))
c                i_dry = nint(dry/(vnorm_mass*dpy))
c                i_drz = nint(drz/(vnorm_mass*dpz))
c                write(*,856)'I,j, j1, j2, drx, dry, drz, az  ',
c     &          I,j, j1, j1, i_drx, i_dry, i_drz, az(i), d_az 
c             elseif(j.eq.i_particleCheck)then   !elseif(j.eq.5661)then
c                i_drx = nint(drx/(vnorm_mass*dpx))
c                i_dry = nint(dry/(vnorm_mass*dpy))
c                i_drz = nint(drz/(vnorm_mass*dpz))
c                write(*,856)'i,J, j1, j2,-drx,-dry,-drz, az  ',
c     &          i,J, j1, j1,-i_drx,-i_dry,-i_drz, az(j), d_az 
cc                if(i.eq.265387)then
cc                  print*,'i, J ',i,j
cc                  print*,'xp(I), zp(I) ',xp(i),zp(i)
cc                  print*,'xp(j), zp(j) ',xp(j),zp(j)
cc                  print*,'drx          ', drx
cc                  print*,'drz          ', drz
cc                  print*,'rr2          ', rr2
cc                  print*,'fourh2       ', fourh2
cc                  print*,'qq=rad/h     ', sqrt(rr2)/h
cc                endif
c             endif           
c           endif
c856        format(a23,2i8,2i6,3i3,2(' ',e24.16))
c           !-------------------------
           
           endif ! if q<2
          enddo ! Box jj
         enddo ! Box ii
641     endif  ! Box jj is not empty
      enddo   ! Kind of particle
      
      end
     