c    "Copyright 2010 Prof. Robert Dalrymple, Prof. M. Gomez Gesteira, Dr Benedict Rogers, 
c     Dr Alejandro Crespo, Dr Muthukumar Narayanaswamy, Dr Shan Zou, Dr Andrea Panizzo "
c
c    This file is part of SPHYSICS.
c
c    SPHYSICS is free software; you can redistribute it and/or modify
c    it under the terms of the GNU General Public License as published by
c    the Free Software Foundation; either version 3 of the License, or
c    (at your option) any later version.
c
c    SPHYSICS is distributed in the hope that it will be useful,
c    but WITHOUT ANY WARRANTY; without even the implied warranty of
c    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
c    GNU General Public License for more details.
c
c    You should have received a copy of the GNU General Public License
c    along with this program.  If not, see <http://www.gnu.org/licenses/>.

c
c*********************************************************************
c   !BDR 04/09/06
c   This subroutine performs the following operations:
c
c     (i)   Checks to see if particles are leaving the subdomain of 
c           the processor (i.e. the rank)
c     (ii)  Transfers the particles and all their properties to the
c           adjacent process
c     (iii) On each processor, the ip index system is then reorganised
c           for the beginning of the next timestep
c
c     Note, this subroutine re-organises the index pointer system but does not
c     reorganise the data array
      
      subroutine identify_Ps_for_Exporting_toBelow
      
      include 'common.3D'


c
!       itime_check = 10e8   !0   !10e8
! 
!       if(rank_id.ge.0.and.itime.ge.itime_check)then
!         print*
!         print*,'BEGINNING OF identify_Ps_for_Exporting_toBelow.f'
!         print*,'itime, rank_id ',itime, rank_id
!         print*,'nb_local ',nb_local
!         print*,'np_local ',np_local
!       end if

      nbf_in_lz_2 = 0
      nbfm_in_lz_2  = 0
      nbfreem_in_lz_2  = 0
      nb_in_lz_2   = 0
      nFP_in_lz_2  = 0
      nbf_in_lz_ncz_localminus1 = 0
      nbfm_in_lz_ncz_localminus1  = 0
      nbfreem_in_lz_ncz_localminus1  = 0
      nb_in_lz_ncz_localminus1   = 0
      nFP_in_lz_ncz_localminus1  = 0

      do kind1=1,2
      
      if(kind1.eq.1)then
        Ntype_check = 3   !Check BPs in order of 1=Fixed BPs; 2=Forced Moving BPs; 3=Free Moving BPs
      else
        Ntype_check = 1
      endif
      
      do i_typeCheck = 1,Ntype_check
      
      do llz = 1,1   !do llx = 1,2
        if(llz.eq.1)then
          lz = 2          !Column to be exported Belowwards
        else
          lz = ncz_local-1  !Column to be exported Abovewards
        end if
      
        do lx = 1,ncx_local
        do ly = 1,ncy_local
      
          j1 = lx + (ly-1)*ncx_local + (lz-1)*nsheet_local
      
          if(nc(j1,kind1).gt.0)then
            
            do ii=1,nc(j1,kind1)
              i = ibox(j1,kind1,ii)  !i here is the absolute position in the array, e.g. xp(i)
                                     !i IS NOT the sequential index which points to position i
            
              !if(rank_id.eq.0)then
              !  print*
              !  print*,'rank_id, itime ',rank_id, itime
              !  print*,'kind1 ',kind1
              !  print*,'llx, lx, ly', llx, lx, ly
              !  print*,'Box j1 ',j1
              !  print*,'nc(j1,kind1) ', nc(j1,kind1)
              !  print*,'is1, ie1', is1,ie1
              !  print*,'ii, i = ip(ii,kind1)',ii,i
              !  print*
              !end if

              if(kind1.eq.2.or.i_particleType(i).eq.i_typeCheck)then

                if(llz.eq.1)then
                  if(i_particleType(i).lt.4)then          !Boundary Particle
                    if(i_particleType(i).eq.1)then        !Fixed BP 
                      nbf_in_lz_2 = nbf_in_lz_2 + 1
                      nb_in_lz_2 = nb_in_lz_2 + 1
                      iBP_going_Below_1(nbf_in_lz_2) = i
                    else if(i_particleType(i).eq.2)then   !Forced Moving BP
                      nbfm_in_lz_2  = nbfm_in_lz_2  + 1
                      nb_in_lz_2 = nb_in_lz_2 + 1
                      iBP_going_Below_2(nbfm_in_lz_2) = i
                    else if(i_particleType(i).eq.3)then    !Free moving BP
                      nbfreem_in_lz_2  = nbfreem_in_lz_2  + 1
                      nb_in_lz_2 = nb_in_lz_2 + 1
                      iBP_going_Below_3(nbfreem_in_lz_2) = i
                    end if                
                    iBP_going_Below(nb_in_lz_2) = i
                  else if(i_particleType(i).eq.4)then      !Fluid Particle
                    nFP_in_lz_2  = nFP_in_lz_2   + 1
                    iFP_going_Below(nFP_in_lz_2) = i
                  else
                    print*
                    print*,'problem: identify_Ps_for_Exporting_toBelow!'
                    print*,'i.NOT.le.nb_local.OR.i.NOT.le.np_local'
                    print*,'itime ',itime
                    print*,'rank_id, ncx_local, ncy_local, ncz_local',
     &                      rank_id, ncx_local, ncy_local, ncz_local
                    print*,'nbf_local ',nbf_local
                    print*,'nbfreem_local  ', nbfreem_local
                    print*,'nb_local   ',nb_local
                    print*,'np_local   ', np_local
                    print*,'kind1 ',kind1
                    print*,'llz, lz ', llz, lz
                    print*,'Box j1 ',j1
                    print*,'nc(j1,kind1) ', nc(j1,kind1)
                    !print*,'is1, ie1', is1,ie1
                    print*,'ii, i = ip(ii,kind1)',ii,i
                    do iii=1,nc(j1,kind1)   !is1,ie1
                      print*,'iii, i, xp(i), yp(i), zp(i)',
     &                        iii, i, xp(i), yp(i), zp(i)
                    end do
                    call MPI_ABORT(MPI_COMM_WORLD, ierr)
                    stop
                  end if
                else if(llz.eq.2)then 
                  if(i_particleType(i).lt.4)then          !Boundary Particle
                    if(i_particleType(i).eq.1)then        !Fixed BP
                      nbf_in_lz_ncz_localminus1=
     &                nbf_in_lz_ncz_localminus1+1
                      nb_in_lz_ncz_localminus1 = 
     &                nb_in_lz_ncz_localminus1 +1
                    else if(i_particleType(i).eq.2)then   !Forced Moving BP
                      nbfm_in_lz_ncz_localminus1 = 
     &                nbfm_in_lz_ncz_localminus1+1
                      nb_in_lz_ncz_localminus1 = 
     &                nb_in_lz_ncz_localminus1 +1
                    else if(i_particleType(i).eq.3)then   !Free moving BP
                      nbfreem_in_lz_ncz_localminus1 = 
     &                nbfreem_in_lz_ncz_localminus1+1
                      nb_in_lz_ncz_localminus1 = 
     &                nb_in_lz_ncz_localminus1 +1
                    end if
                    iBP_going_Above(nb_in_lz_ncz_localminus1) = i
                  else if(i_particleType(i).eq.4)then     !Fluid Particle
                    nFP_in_lz_ncz_localminus1 = 
     &              nFP_in_lz_ncz_localminus1 +1
                    iFP_going_Above(nFP_in_lz_ncz_localminus1) = i
                    !if(rank_id.eq.0)then
                    !  print*
                    !  print*,'Incrementing nFP_in_lz_ncz_localminus1' 
                    !  print*,'rank_id ',rank_id
                    !  print*,'llx, lx, ly', llx, lx, ly
                    !  print*,'ii, i = ip(ii,kind1)',ii,i
                    !  print*,'xp(i), yp(i) ',xp(i), yp(i) 
                    !end if
                  else
                    print*
                    print*,'problem: identify_Ps_for_Exporting_toBelow!'
                    print*,'i.NOT.le.nb_local.OR.i.NOT.le.np_local'
                    print*,'itime ',itime
                    print*,'rank_id, ncx_local, ncy_local, ncz_local',
     &                      rank_id, ncx_local, ncy_local, ncz_local
                    print*,'nbf_local ',nbf_local
                    print*,'nbfreem_local  ', nbfreem_local
                    print*,'nb_local   ',nb_local
                    print*,'np_local   ', np_local
                    print*,'kind1 ',kind1
                    print*,'llz, lz ', llz, lz
                    print*,'Box j1 ',j1
                    print*,'nc(j1,kind1) ', nc(j1,kind1)
c                    print*,'is1, ie1', is1,ie1
c                    print*,'ii, i = ip(ii,kind1)',ii,i
c                    do iii=1,nc(j1,kind1)
c                      i_i = ibox(j1,kind1,iii)   !ip(ii,kind1)
c                        print*,'iii, i_i, xp(i_i), yp(i_i), zp(i_i) ',
c     &                        iii, i_i, xp(i_i), yp(i_i), zp(i_i)
c                    end do
c                    print*,'ii, i = ip(ii,kind1)',ii,i
                    do iii=1,nc(j1,kind1)   !is1,ie1
                      print*,'iii, i, xp(i), yp(i), zp(i)',
     &                        iii, i, xp(i), yp(i), zp(i)
                    end do
                    call MPI_ABORT(MPI_COMM_WORLD, ierr)
                    stop
                  end if
                end if
              
              endif    !End of:  if(kind1.eq.2.or.i_particleType(i).eq.i_typeCheck)then  
              
            end do

          end if    !End of:  if(nc(j1,kind).gt.0)then

        end do   !End of:  do ly = 1,ncy_local
        end do   !End of:  do lx = 1,ncx_local

      end do
      
      enddo   !End of:  do i_typeCheck = 1,3
      
      end do
      
      !end if
      !stop
      
      nbf0_in_lz_2     = nbf_in_lz_2
      nbfm0_in_lz_2    = nbfm_in_lz_2
      nbfreem0_in_lz_2 = nbfreem_in_lz_2
      nFP0_in_lz_2     = nFP_in_lz_2

      !Make numbers absolute in reference to start of local list
      !This makes the nomenclature consistent
      nbfm_in_lz_2 = nbf_in_lz_2 + nbfm_in_lz_2
      nbfreem_in_lz_2 = nbfm_in_lz_2 + nbfreem_in_lz_2
      np_in_lz_2 = nFP_in_lz_2  + nb_in_lz_2
      nbfm_in_lz_ncz_localminus1 = nbf_in_lz_ncz_localminus1 
     &                          + nbfm_in_lz_ncz_localminus1
      nbfreem_in_lz_ncz_localminus1 = nbfm_in_lz_ncz_localminus1 
     &                            + nbfreem_in_lz_ncz_localminus1
      np_in_lz_ncz_localminus1 = nFP_in_lz_ncz_localminus1
     &                         + nb_in_lz_ncz_localminus1
      
!       !- Diagnostic Info -
!       if(itime.ge.itime_check)then
!         if(rank_id.eq.0)then   !if(rank_id.eq.-10)then
!           print*,'Checking Particles to be exported'
!         end if
!         do i=0,n_procs-1
!           if(i.eq.rank_id)then   !if(i.eq.-rank_id-1)then   !if(i.eq.rank_id)then
!             print*
!             print*,'rank_id ', rank_id
!             print*,'nbf_in_lz_2 ',nbf_in_lz_2
!             print*,'nbfreem_in_lz_2  ',nbfreem_in_lz_2
!             print*,'nb_in_lz_2   ',nb_in_lz_2
!             print*,'nFP_in_lz_2  ',nFP_in_lz_2
!             print*,'np_in_lz_2   ',np_in_lz_2
!             print*,'nbf_in_lz_ncz_localminus1 ',
!      &              nbf_in_lz_ncz_localminus1
!             print*,'nbfreem_in_lz_ncz_localminus1  ',
!      &              nbfreem_in_lz_ncz_localminus1
!             print*,'nb_in_lz_ncz_localminus1   ',
!      &              nb_in_lz_ncz_localminus1
!             print*,'nFP_in_lz_ncz_localminus1  ',
!      &              nFP_in_lz_ncz_localminus1
!             print*,'np_in_lz_ncz_localminus1   ',
!      &              np_in_lz_ncz_localminus1
!           end if
!           call MPI_BARRIER(grid_comm, ierr)   !Synchronise all processes to this point
!         end do
!       end if
      
      !stop
      
      nb_going_Above = nb_in_lz_ncz_localminus1
      np_going_Above = np_in_lz_ncz_localminus1
      nb_going_Below = nb_in_lz_2
      np_going_Below = np_in_lz_2
      
      !-- Zero non-communicating processes --
      if(rank_idz.eq.0
     &   .and.(.NOT.iCartTop_periodicity(3)))then
        nb_going_Below = 0
        np_going_Below = 0
        if(n_procs_z.eq.1)then
          nb_going_Above = 0
          np_going_Above = 0
        end if
      else if(rank_idz.eq.n_procs_z-1
     &   .and.(.NOT.iCartTop_periodicity(3)))then
        nb_going_Above = 0
        np_going_Above = 0
        if(n_procs_z.eq.1)then
          nb_going_Below = 0
          np_going_Below = 0
        end if
      end if

!      !- Diagnostic Info -
!       if(itime.ge.itime_check)then
!         if(rank_id.eq.-10)then
!           print*
!           print*,'At END of identify_Ps_for_Exporting_toBelow'
!           print*,'Checking Particles to be exported'
!           print*,'rank_id, itime ',rank_id, itime
!           print*,'nb_going_Below  ',nb_going_Below
!           print*,'np_going_Below  ',np_going_Below
!           print*
!           print*,'BPs going Below'
!           do ii = 1,nb_going_Below
!             i = iBP_going_Below(ii)
!             print*,'ii, i, xp, yp, zp ',ii, i, xp(i), yp(i), zp(i)
!           end do
!           print*,'FPs going Below'
!           do ii = nb_going_Below+1,nb_going_Below
!             i = iFP_going_Below(ii-nb_going_Below)
!             print*,'ii, i, xp, yp, zp ',ii, i, xp(i), yp(i), zp(i)
!           end do
!         end if
!       end if
      
c      if(rank_id.eq.2.and.itime.ge.0)then
c        print*
c        print*,'END OF identify_Ps_for_Exporting_toBelow.f'
c        print*,'itime, rank_id ',itime, rank_id
c        !print*,'nb_going_Above ',nb_going_Above
c        !print*,'np_going_Above ',np_going_Above
c        print*,'nb_going_Below ',nb_going_Below
c        print*,'np_going_Below ',np_going_Below
c        print*,'nbf_in_lz_2       ',nbf_in_lz_2
c        print*,'nbfm_in_lz_2      ',nbfm_in_lz_2
c        print*,'nbfreem_in_lz_2   ',nbfreem_in_lz_2
c        print*,'nb_in_lz_2        ',nb_in_lz_2
c        print*,'np_in_lz_2        ',np_in_lz_2
c        print*,'np_local ',np_local
c      end if

      !stop

      end subroutine
