c    "Copyright 2010 Prof. Robert Dalrymple, Prof. M. Gomez Gesteira, Dr Benedict Rogers, 
c     Dr Alejandro Crespo, Dr Muthukumar Narayanaswamy, Dr Shan Zou, Dr Andrea Panizzo "
c
c    This file is part of SPHYSICS.
c
c    SPHYSICS is free software; you can redistribute it and/or modify
c    it under the terms of the GNU General Public License as published by
c    the Free Software Foundation; either version 3 of the License, or
c    (at your option) any later version.
c
c    SPHYSICS is distributed in the hope that it will be useful,
c    but WITHOUT ANY WARRANTY; without even the implied warranty of
c    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
c    GNU General Public License for more details.
c
c    You should have received a copy of the GNU General Public License
c    along with this program.  If not, see <http://www.gnu.org/licenses/>.


      subroutine monaghanBC(iip,jjp,drx,dry,drz,dot,
     &                      duxp,duyp,duzp,
     &                      fxbp,fybp,fzbp)
c
      include 'common.3D'
 
c     Boundary condition JWPCOE, 32
c
c     iip is a WaterParticle (WP)
c     jjp is a BoundaryParticle (BP)
c
c     drx = xp(WP) - xp(BedP)
c     dry = yp(WP) - yp(BedP)
c     drz = zp(WP) - zp(BedP)

c      Projection of WP position onto normal
       pdot3 = drx*xnb(jjp)+dry*ynb(jjp)+drz*znb(jjp)

c       if(pdot3.gt.0.0)then  !ensures that WP is on correct side of boundary
c
c         xpp = magnitude(vec(dr) x vec(n))   cross-product
c
c         xpp=(drx*ynb(jjp)-dry*xnb(jjp))


c        Projection of WP on plane of boundary
c
c        r(BP->WP) = [drx,dry,drz]
c
c        r(BP->WP projection) = r(BP->WP) -[r(BP->WP).n/(|n|.|n|)]n
c
c        normal to plane = n
c        magnitude of normal = |n| = 1

         drx_proj = drx - (pdot3)*xnb(jjp)
         dry_proj = dry - (pdot3)*ynb(jjp)
         drz_proj = drz - (pdot3)*znb(jjp)

c        Projection of plane-projected vector r(BP->WP projection) onto tangent

         xpp = drx_proj*xtb(jjp)+dry_proj*ytb(jjp)+drz_proj*ztb(jjp)
         ypp = drx_proj*xsb(jjp)+dry_proj*ysb(jjp)+drz_proj*zsb(jjp)

         if(xpp.gt.0.0)then
           deltappt = deltaptb(jjp,2)
         else
           deltappt = deltaptb(jjp,1)
         end if
         if(ypp.gt.0.0)then
           deltapps = deltapsb(jjp,2)
         else
           deltapps = deltapsb(jjp,1)
         end if

         !- Correction for particle exactly deltappt from another particle -
         g_dot_t = grx*xtb(jjp) + gry*ytb(jjp) + grz*ztb(jjp)
         xpp_orig = xpp
         if(.NOT.(abs(xpp).lt.deltappt))then  !i.e. >=deltappt
           if(g_dot_t.gt.0.0)then
             if(xpp.gt.0.0)then
               xpp = 0.9999*xpp
             endif
           else   !if(g_dot_t.lt.0.0)then
             if(xpp.lt.0.0)then
               xpp = 0.9999*xpp
             endif
           endif
         elseif(.NOT.(abs(xpp).gt.0.0))then  !i.e. ==0.0
           if(g_dot_t.gt.0.0)then
             xpp = xpp - 0.0001*deltappt
           else   !if(g_dot_t.lt.0.0)then
             xpp = xpp + 0.0001*deltappt
           endif
         endif
         !- Correction for particle exactly deltapps from another particle -
         g_dot_s = grx*xsb(jjp) + gry*ysb(jjp) + grz*zsb(jjp)
         ypp_orig = ypp
         if(.NOT.(abs(ypp).lt.deltapps))then  !i.e. >=deltapps
           if(g_dot_s.gt.0.0)then
             if(ypp.gt.0.0)then
               ypp = 0.9999*ypp
             endif
           else   !if(g_dot_s.lt.0.0)then
             if(ypp.lt.0.0)then
               ypp = 0.9999*ypp
             endif
           endif
         elseif(.NOT.(abs(ypp).gt.0.0))then  !i.e. ==0.0
           if(g_dot_s.gt.0.0)then
             ypp = ypp - 0.0001*deltapps
           else   !if(g_dot_s.lt.0.0)then
             ypp = ypp + 0.0001*deltapps
           endif
         endif
         
         Prx = abs(xpp)/deltappt    
         Pry = abs(ypp)/deltapps      
         facp = 0.0
         f_tb = 0.0
         f_fb = 0.0
         chi_X = 0.0
         chi_Y = 0.0   
         Q_fn = 0.0
         i_CornerParticle = 0
         
         !normalized distance from wall
         if(pdot3.gt.0.0)then ! particle inside
             q=pdot3 * one_over_2h           !normalized distance from wall
         else if(pdot3.lt.0.0)then !Particle ouside 
             q=0.2*abs(pdot3) * one_over_2h  !normalized distance from wall  
         else !Particle on the boundary
             q=h*(1.0e-3) * one_over_2h      !correcting for zero distance from wall
         end if
                  
         if(q.lt.1.0.and.pdot3.gt.0.0.and.
     &      abs(xpp).lt.deltappt.and.abs(ypp).lt.deltapps)then
     

c            u_normal < 0 if particles are approaching
c            u_normal > 0 if particles are moving apart
      
           q_n = q  !/vnorm_mass
           Q_fn = (1.0-q_n)/sqrt(q_n)   !Q_fn = (1.0-q)/sqrt(q)
           u_normal = duxp*xnb(jjp) + duyp*ynb(jjp) + duzp*znb(jjp)
           chi_X = (1.0 - Prx)
           chi_Y = (1.0 - Pry)              
           q_r = q
           xtb_r = xtb(jjp)
           ytb_r = ytb(jjp)
           ztb_r = ztb(jjp)
           xsb_r = xsb(jjp)
           ysb_r = ysb(jjp)
           zsb_r = zsb(jjp)
             
           if(chi_X.gt.0.0.and.chi_Y.gt.0.0.and.Q_fn.gt.0.0)then
             !c0 = coeff.Vmax (Vmax = sqrt(g.h_SWL); coef = beta_coef.10, beta_coef2 = beta_coef**2
             !beta_coef2 = Factor that B is changed by 
             if(u_normal.gt.0.0)then
              App=(0.01*cs(iip)*cs(iip)*one_over_beta_coef2
     &                -cs(iip)*u_normal*one_over_beta_coef)*one_over_h
             else
              App=0.01*(cs(iip)*cs(iip)*one_over_beta_coef2)*one_over_h    
             endif

	
c          --- Depth Correction ---            
             eps_min = 0.02
             epsilon_Z = abs(1.0 - (zp(jjp)/h_SWL))
             dyn_head =(up(jjp)*up(jjp) + vp(jjp)*vp(jjp)
     &                + wp(jjp)*wp(jjp))  /(2.0*abs(grz)*h_SWL)
           
             if(epsilon_Z.gt.1.0-eps_min)then !Close to the bottom (No correction )
                epsilon_Z = 1.0
             elseif(zp(iip).gt.h_SWL)then !over still water level          
                epsilon_Z = eps_min + dyn_head
             else                 !under still water level & far from bottom
                epsilon_Z = eps_min + epsilon_Z + dyn_head
             endif

c          --- Velocity Correction ---
             if(u_normal.gt.0.0)then
                epsilon_dyn = 0.0
             else 
                epsilon_dyn = min(1.0,40.*abs(u_normal)/cs0)     !factor 20 needed                                                       ! to keep scheme stable                                                                                  ! with Monaghan BC wavemaker
             endif

             Appc = App*(epsilon_Z + epsilon_dyn) !Force correction
             
             fmax = Appc*Q_fn  !*vnorm_mass
             facp = chi_X*chi_Y*fmax
             
             fxbp = xnb(jjp)*facp
             fybp = ynb(jjp)*facp
             fzbp = znb(jjp)*facp

             !- Angular Momentum Correction -
c               g_dot_t = grx*xtb_r + gry*ytb_r + grz*ztb_r
c               g_dot_s = grx*xsb_r + gry*ysb_r + grz*zsb_r
c               g_dot_t_normalised = abs(g_dot_t + g_dot_s)/abs(grz)
c               !deltapp = sqrt(deltappt**2 + deltapps**2)
c               deltapp = sqrt(2.0)*deltappt  !Temp for debugging
c               f_tb =  rootTwo_alpha_h_over_4*g_dot_t_normalised*fmax*
c     &                   (0.5*deltapp*one_over_r_ij_dot_n)
c     &                   *one_over_Q_of_deltax
               f_tb =  two_alpha_h_over_4*fmax  !*one_over_r_ij_dot_n
     &                   *one_over_Q_of_deltax
c               g_dot_t_normalised = abs(g_dot_t)/abs(grz)
c               g_dot_s_normalised = abs(g_dot_s)/abs(grz)

               !- Correction for WP near bottom corner or surface
               h_SWL_minus_zp = h_SWL - zp(iip)
               !rNum_h = 3.0*h
               !one_over_rNum_h = 1.0/rNum_h
               f_tb_orig = f_tb
               if(h_SWL_minus_zp.lt.0.0)then  !WP is above h_SWL
                 f_tb = 0.0  !*f_tb
               elseif(h_SWL_minus_zp.lt.rNum_h)then
                 epsilon_Z_AMC = h_SWL_minus_zp * one_over_rNum_h
                 if(epsilon_Z_AMC.lt.0.90)then
                   epsilon_Z_AMC = (epsilon_Z_AMC + 0.10)
                 endif
                 f_tb = f_tb * epsilon_Z_AMC
               !elseif(zp(iip).lt.rNum_h)then  !i.e. less than 2h above the bottom
               !  epsilon_Z_AMC = zp(iip) * one_over_rNum_h
               !  f_tb = f_tb * epsilon_Z_AMC
               endif
               
               if(xpp.gt.0.0)then
                 sign_rij_dot_tj =   1.0
               else
                 sign_rij_dot_tj =  -1.0
               end if
               if(ypp.gt.0.0)then
                 sign_rij_dot_sj =   1.0
               else
                 sign_rij_dot_sj =  -1.0
               end if
               f_tb_t = f_tb !* g_dot_t_normalised
     &                       * sign_rij_dot_tj !* deltappt
               f_tb_s = f_tb !* g_dot_s_normalised
     &                       * sign_rij_dot_sj !* deltapps
               
               fxbp = fxbp + xtb_r*f_tb_t + xsb_r*f_tb_s
               fybp = fybp + ytb_r*f_tb_t + ysb_r*f_tb_s
               fzbp = fzbp + ztb_r*f_tb_t + zsb_r*f_tb_s
               
           else
             fxbp = 0.0
             fybp = 0.0
             fzbp = 0.0
           endif               
         else
             fxbp = 0.0
             fybp = 0.0
             fzbp = 0.0
         end if

c      !- Debugging -
c      fxbp = 0.0
c      fybp = 0.0
c      fzbp = 0.0

c      if(1.eq.2.and.
c     &   itime.le.1.and. 
ccc     & (iip.eq.3455))then
ccc     & (iip.eq.6302))then
ccc     & (iip.eq.6367))then
c     & (iip.eq.2531.and.rank_id.eq.0))then  !16585
ccc         if(q.lt.1.0.and.q.gt.0.0.and.
ccc     &      abs(xpp).lt.deltappt)then
cc         if(q.lt.1.0.and.q.gt.0.0)then
c         if(Prx.lt.1.01.and.Pry.lt.1.01)then
c               print*
c               print*,'itime ',itime
c               print*,'rhop(iip), cs(iip) ',rhop(iip), cs(iip)
c               print*,'iip, xp, yp, zp ',iip,xp(iip),yp(iip),zp(iip)
c               print*,'jjp, xp, yp, zp ',jjp,xp(jjp),yp(jjp),zp(jjp)
c               print*,'drx, dry, drz  ',drx, dry, drz
c               print*,'drx_proj, dry_proj, drz_proj ',
c     &                 drx_proj, dry_proj, drz_proj
c               print*,'xpp_orig, xpp, deltappt ',xpp_orig, xpp, deltappt
c               print*,'ypp_orig, ypp, deltapps ',ypp_orig, ypp, deltapps
c               print*,'abs(xpp).lt.deltappt ',abs(xpp).lt.deltappt
c               print*,'abs(ypp).lt.deltapps ',abs(ypp).lt.deltapps
c               print*,'.NOT.(abs(xpp_orig).lt.deltappt) ',
c     &                 .NOT.(abs(xpp_orig).lt.deltappt)
c               print*,'.NOT.(abs(ypp_orig).lt.deltapps) ',
c     &                 .NOT.(abs(ypp_orig).lt.deltapps)
c               print*,'.NOT.(abs(xpp_orig).gt.0.0) ',
c     &                 .NOT.(abs(xpp_orig).gt.0.0)
c               print*,'.NOT.(abs(ypp_orig).gt.0.0) ',
c     &                 .NOT.(abs(ypp_orig).gt.0.0)
c               print*,'q  , q_r, pdot3  ',q, q_r, pdot3
c               print*,'up, vp, wp(iip)   ',up(iip),vp(iip),wp(iip)
c               print*,'up, vp, wp(jjp)   ',up(jjp),vp(jjp),wp(jjp)
c               print*,'Particle jjp ',jjp
c               print*,'xnb, ynb, znb ',
c     &                 xnb(jjp),ynb(jjp),znb(jjp)
c               print*,'xtb, ytb, ztb ',
c     &                 xtb(jjp),ytb(jjp),ztb(jjp)
c               print*,'xsb, ysb, zsb ',
c     &                 xsb(jjp),ysb(jjp),zsb(jjp)
c               print*,'xtb_r, ytb_r, ztb_r ',xtb_r,ytb_r,ztb_r
c               print*,'xsb_r, ysb_r, zsb_r ',xsb_r,ysb_r,zsb_r
c               print*,'g_dot_t, g_dot_s   ',g_dot_t, g_dot_s
c               print*,'g_dot_t_normalised ',g_dot_t_normalised
c               print*,'h_SWL, rNum_h ',
c     &                 h_SWL, rNum_h
c               print*,'h_SWL_minus_zp, one_over_rNum_h ',
c     &                 h_SWL_minus_zp, one_over_rNum_h
cc               print*,'rootTwo_alpha_h_over_4 ',rootTwo_alpha_h_over_4
cc               print*,'deltappt, deltapps, deltapp ',
cc     &                 deltappt, deltapps, deltapp
c               print*,'two_alpha_h_over_4 ',two_alpha_h_over_4
c               print*,'deltappt, deltapps ',
c     &                 deltappt, deltapps
c               print*,'one_over_r_ij_dot_n  ',one_over_r_ij_dot_n
c               print*,'one_over_Q_of_deltax ',one_over_Q_of_deltax
c               print*,'f_tb_orig, epsilon_Z_AMC  ',
c     &                 f_tb_orig, epsilon_Z_AMC
c               print*,'u_normal    ',u_normal
c               print*,'u_parallel  ',u_parallel
c               print*,'Prx,chi_X ',Prx,chi_X
c               print*,'Pry,chi_Y ',Pry,chi_Y
c               print*,'q, Q_fn   ',q, Q_fn
c               print*,'fmax, facp ',fmax, facp
c               print*,'Appc, vnorm_mass    ',Appc, vnorm_mass
c               print*,'App, epsilon_Z, epsilon_dyn ',
c     &                 App, epsilon_Z, epsilon_dyn
c               print*,'fxbp, fybp, fzbp  ',fxbp, fybp, fzbp
c               print*,'facp, f_tb  ',facp,f_tb
c               print*,'f_tb_t, f_tb_s  ',f_tb_t, f_tb_s
c               print*,'sign_rij_dot_tj, sign_rij_dot_sj  ',
c     &                 sign_rij_dot_tj, sign_rij_dot_sj
c               print*,'f_tb_orig, epsilon_Z_AMC  ',
c     &                 f_tb_orig, epsilon_Z_AMC
c               print*,'xtb_r*f_tb_t= ',xtb_r*f_tb_t,
c     &             ' , xsb_r*f_tb_s= ',xsb_r*f_tb_s
c               print*,'ytb_r*f_tb_t= ',ytb_r*f_tb_t,
c     &             ' , ysb_r*f_tb_s= ',ysb_r*f_tb_s
c               print*,'ztb_r*f_tb_t= ',ztb_r*f_tb_t,
c     &             ' , zsb_r*f_tb_s= ',zsb_r*f_tb_s
cc               !print*,'sign(1.0,u_parallel) ',sign(1.0,u_parallel)
cc               print*,'sign_u_parallel      ',sign_u_parallel
cc               print*,'xnb(jjp)*,znb(jjp)*facp ',
cc     &                 xnb(jjp)*facp, znb(jjp)*facp
cc               print*,'xtb_r*f_tb, ztb_r*f_tb ',
cc     &                 xtb_r*f_tb, ztb_r*f_tb
cc               print*,'jjp_absolute, j_minus1, j_plus1 ',
cc     &                 jjp_absolute, j_minus1, j_plus1
cc               print*,'i_CornerParticle ',i_CornerParticle
cc               if(i_CornerParticle.gt.0)then
cc                 print*,'drp, one_over_drp ',drp, one_over_drp
cc                 print*,'1./drp, h_over_1000 ',1.0/drp, h_over_1000
cc                 print*,
cc     &               ' jjp_minus1, jjp_plus1 ',
cc     &                 jjp_minus1, jjp_plus1
cc                 print*,'vn1_dot_rij, vn1_dot_n2 ',
cc     &                   vn1_dot_rij, vn1_dot_n2
cc               endif
cc         print*
c         !read(*,*)
c        end if
c        if(itime.eq.1)stop
c      end if
 

      end	