c    "Copyright 2010 Prof. Robert Dalrymple, Prof. M. Gomez Gesteira, Dr Benedict Rogers, 
c     Dr Alejandro Crespo, Dr Muthukumar Narayanaswamy, Dr Shan Zou, Dr Andrea Panizzo "
c
c    This file is part of SPHYSICS.
c
c    SPHYSICS is free software; you can redistribute it and/or modify
c    it under the terms of the GNU General Public License as published by
c    the Free Software Foundation; either version 3 of the License, or
c    (at your option) any later version.
c
c    SPHYSICS is distributed in the hope that it will be useful,
c    but WITHOUT ANY WARRANTY; without even the implied warranty of
c    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
c    GNU General Public License for more details.
c
c    You should have received a copy of the GNU General Public License
c    along with this program.  If not, see <http://www.gnu.org/licenses/>.


      subroutine monaghanBC(iip,jjp,drx,dry,drz,dot,
     &                      duxp,duyp,duzp,
     &                      fxbp,fybp,fzbp)
c
      include 'common.3D'
      id_num_FB_i = 0
      id_num_FB_j = 0
c     Boundary condition JWPCOE, 32
c
c     iip is a WaterParticle (WP)
c     jjp is a BoundaryParticle (BP)
c
c     drx = xp(WP) - xp(BedP)
c     drz = yp(WP) - yp(BedP)
c     drz = zp(WP) - zp(BedP)

c      Projection of WP position onto normal
       pdot3 = drx*xnb(jjp)+dry*ynb(jjp)+drz*znb(jjp)

c       if(pdot3.gt.0.0)then  !ensures that WP is on correct side of boundary
c
c         xpp = magnitude(vec(dr) x vec(n))   cross-product
c
c         xpp=(drx*ynb(jjp)-drz*xnb(jjp))


c        Projection of WP on plane of boundary
c
c        r(BP->WP) = [drx,drz,drz]
c
c        r(BP->WP projection) = r(BP->WP) -[r(BP->WP).n/(|n|.|n|)]n
c
c        normal to plane = n
c        magnitude of normal = |n| = 1

         drx_proj = drx - (pdot3)*xnb(jjp)
         dry_proj = dry - (pdot3)*ynb(jjp)
         drz_proj = drz - (pdot3)*znb(jjp)

c        Projection of plane-projected vector r(BP->WP projection) onto tangent

         xpp = drx_proj*xtb(jjp)+dry_proj*ytb(jjp)+drz_proj*ztb(jjp)
         ypp = drx_proj*xsb(jjp)+dry_proj*ysb(jjp)+drz_proj*zsb(jjp)

         if(xpp.gt.0.0)then
           deltappt = deltaptb(jjp,2)
         else
           deltappt = deltaptb(jjp,1)
         end if
         if(ypp.gt.0.0)then
           deltapps = deltapsb(jjp,2)
         else
           deltapps = deltapsb(jjp,1)
         end if

         !- Correction for particle exactly deltappt from another particle -
         g_dot_t = grx*xtb(jjp) + gry*ytb(jjp) + grz*ztb(jjp)
         xpp_orig = xpp
         if(.NOT.(abs(xpp).lt.deltappt))then  !i.e. >=deltappt
           if(g_dot_t.gt.0.0)then
             if(xpp.gt.0.0)then
               xpp = 0.9999*xpp
             endif
           else   !if(g_dot_t.lt.0.0)then
             if(xpp.lt.0.0)then
               xpp = 0.9999*xpp
             endif
           endif
         elseif(.NOT.(abs(xpp).gt.0.0))then  !i.e. ==0.0
           if(g_dot_t.gt.0.0)then
             xpp = xpp - 0.0001*deltappt
           else   !if(g_dot_t.lt.0.0)then
             xpp = xpp + 0.0001*deltappt
           endif
         endif
         !- Correction for particle exactly deltapps from another particle -
         g_dot_s = grx*xsb(jjp) + gry*ysb(jjp) + grz*zsb(jjp)
         ypp_orig = ypp
         if(.NOT.(abs(ypp).lt.deltapps))then  !i.e. >=deltapps
           if(g_dot_s.gt.0.0)then
             if(ypp.gt.0.0)then
               ypp = 0.9999*ypp
             endif
           else   !if(g_dot_s.lt.0.0)then
             if(ypp.lt.0.0)then
               ypp = 0.9999*ypp
             endif
           endif
         elseif(.NOT.(abs(ypp).gt.0.0))then  !i.e. ==0.0
           if(g_dot_s.gt.0.0)then
             ypp = ypp - 0.0001*deltapps
           else   !if(g_dot_s.lt.0.0)then
             ypp = ypp + 0.0001*deltapps
           endif
         endif

         Prx = abs(xpp)/deltappt    
         Pry = abs(ypp)/deltapps      
         facp = 0.0
         f_tb = 0.0
         f_fb = 0.0
         chi_X = 0.0
         chi_Y = 0.0   
         Q_fn = 0.0
         i_CornerParticle = 0
         
         !normalized distance from wall
         if(pdot3.gt.0.0)then ! particle inside
             q=pdot3 * one_over_2h           !normalized distance from wall
         else if(pdot3.lt.0.0)then !Particle ouside 
             q=0.2*abs(pdot3) * one_over_2h  !normalized distance from wall
         else !Particle on the boundary
             q=h*(1.0e-3) * one_over_2h      !correcting for zero distance from wall
         end if
                  
         i_particleType_iip = i_particleType(iip)
         i_particleType_jjp = i_particleType(jjp)
         if(q.lt.1.0.and.
     &      abs(xpp).lt.deltappt.and.abs(ypp).lt.deltapps)then

           if(i_particleType_iip.lt.4.or.   !If not, Therefore i_particleType_iip.eq.4 (FP)
     &        ((i_particleType_jjp.eq.3.and.pdot3.gt.0.0).or.    !Rigid Body BP
     &         (i_particleType_jjp.lt.3)                      ))then   !Fixed or Forced Motion BPs
     
c            u_normal < 0 if particles are approaching
c            u_normal > 0 if particles are moving apart
      
             q_n = q  !/vnorm_mass
             Q_fn = (1.0-q_n)/sqrt(q_n)   !Q_fn = (1.0-q)/sqrt(q)
             u_normal = duxp*xnb(jjp) + duyp*ynb(jjp) + duzp*znb(jjp)
             chi_X = (1.0 - Prx)
             chi_Y = (1.0 - Pry)              
             q_r = q
             xtb_r = xtb(jjp)
             ytb_r = ytb(jjp)
             ztb_r = ztb(jjp)
             xsb_r = xsb(jjp)
             ysb_r = ysb(jjp)
             zsb_r = zsb(jjp)

             !-- Corner Particle Identification --
             jjp_absolute = iabsolute_BP_Pointer_Index(jjp)
             i_minus1 = iBP_Pointer_Info(jjp_absolute,3)     !Absolute index of i-1 neighbour BP
             i_plus1  = iBP_Pointer_Info(jjp_absolute,4)     !Absolute index of i+1 neighbour BP
             j_minus1 = iBP_Pointer_Info(jjp_absolute,5)     !Absolute index of i-1 neighbour BP
             j_plus1  = iBP_Pointer_Info(jjp_absolute,6)     !Absolute index of i+1 neighbour BP
             
             if(chi_X.gt.0.0.and.chi_Y.gt.0.0.and.Q_fn.gt.0.0)then
               !c0 = coeff.Vmax (Vmax = sqrt(g.h_SWL); coef = beta_coef.10, beta_coef2 = beta_coef**2
               !beta_coef2 = Factor that B is changed by 
               if(i_particleType_iip.eq.4)then   !Normal FP-BP Interaction   !if(iip.gt.nb_local)then  
                 if(u_normal.gt.0.0)then
                   App=(0.01*cs(iip)*cs(iip)*one_over_beta_coef2
     &                -cs(iip)*u_normal*one_over_beta_coef)*one_over_h
                 else
                   App=0.01*(cs(iip)*cs(iip)*one_over_beta_coef2)
     &                     *one_over_h    
                 endif
               else !BP-BP interactions
                 cs_i = cs0
                 if(u_normal.gt.0.0)then    !Particles receding:  u_normal > 0
                   cs2_diff = 0.5*abs(cs_i*cs_i)   !
                   App=(0.01*cs2_diff*one_over_beta_coef2
     &                 -cs_i*u_normal*one_over_beta_coef)*one_over_h
                 else                       !Particles approaching:  u_normal < 0
                   !cs2_diff = 0.5*abs(cs_i*cs_i)   !Set this for elastic bouncing
                   cs2_diff = 0.0   !Set this for elastic bouncing
                   repulsion_factor = 10.0
                   !if(iip.gt.nbfm.and.jjp.gt.nbfm)then
                   !  repulsion_factor = 2.0
                   !endif
                   App=(0.01*cs2_diff*one_over_beta_coef2
     &                 -repulsion_factor*cs_i*u_normal
     &                 *one_over_beta_coef)*one_over_h
                 endif
                 App = App*0.5
                 if(q.lt.0.2.and.
     &            ((jjp_absolute.ne.i_minus1.and.
     &              jjp_absolute.ne.i_plus1      ).or.
     &             (jjp_absolute.ne.j_minus1.and.
     &              jjp_absolute.ne.j_plus1     )))then
                   q_2 = q *q   !Also influences tangential forces!!
                   Q_fn = (1.0-q_2)/sqrt(q_2)   !Q_fn = (1.0-q)/sqrt(q)
                 end if
               endif    !End of: if(i_particleType_iip.eq.4)then

	
c            --- Depth Correction ---            
               eps_min = 0.02
               epsilon_Z = abs(1.0 - (zp(jjp)/h_SWL))
               dyn_head =(up(jjp)*up(jjp) + vp(jjp)*vp(jjp)
     &                   +wp(jjp)*wp(jjp))  /(2.0*abs(grz)*h_SWL)
           
               if(epsilon_Z.gt.1.0-eps_min)then !Close to the bottom (No correction )
                 epsilon_Z = 1.0
               elseif(zp(iip).gt.h_SWL)then !over still water level          
                 epsilon_Z = eps_min + dyn_head
               else                 !under still water level & far from bottom
                 epsilon_Z = eps_min + epsilon_Z + dyn_head
               endif

c            --- Velocity Correction ---
               if(u_normal.gt.0.0)then
                 epsilon_dyn = 0.0
               else 
                 epsilon_dyn = min(1.0,40.*abs(u_normal)/cs0)     !factor 20 needed                                                       ! to keep scheme stable                                                                                  ! with Monaghan BC wavemaker
               endif

               Appc = App*(epsilon_Z + epsilon_dyn) !Force correction

               fmax = Appc*Q_fn  !*vnorm_mass
               facp = chi_X*chi_Y*fmax
             
               fxbp = xnb(jjp)*facp
               fybp = ynb(jjp)*facp
               fzbp = znb(jjp)*facp


               !- Angular Momentum Correction -
               if(i_particleType_iip.eq.4.and.
     &            i_particleType_jjp.lt.4     )then  !Floating Body Version
                 f_tb =  two_alpha_h_over_4*fmax  !*one_over_r_ij_dot_n
     &                     *one_over_Q_of_deltax
                 !- Correction for WP near bottom corner or surface
                 h_SWL_minus_zp = h_SWL - zp(iip)
                 !rNum_h = 3.0*h
                 !one_over_rNum_h = 1.0/rNum_h
                 f_tb_orig = f_tb
                 if(h_SWL_minus_zp.lt.0.0)then  !WP is above h_SWL
                   f_tb = 0.0  !*f_tb
                 elseif(h_SWL_minus_zp.lt.rNum_h)then
                   epsilon_Z_AMC = h_SWL_minus_zp * one_over_rNum_h
                   if(epsilon_Z_AMC.lt.0.90)then
                     epsilon_Z_AMC = (epsilon_Z_AMC + 0.10)
                   endif
                   f_tb = f_tb * epsilon_Z_AMC
                 !elseif(zp(iip).lt.rNum_h)then  !i.e. less than 2h above the bottom
                 !  epsilon_Z_AMC = zp(iip) * one_over_rNum_h
                 !  f_tb = f_tb * epsilon_Z_AMC
                 endif
               
                 if(xpp.gt.0.0)then
                   sign_rij_dot_tj =   1.0
                 else
                   sign_rij_dot_tj =  -1.0
                 end if
                 if(ypp.gt.0.0)then
                   sign_rij_dot_sj =   1.0
                 else
                   sign_rij_dot_sj =  -1.0
                 end if
                 f_tb_t = f_tb * sign_rij_dot_tj
                 f_tb_s = f_tb * sign_rij_dot_sj
               
                 fxbp = fxbp + xtb_r*f_tb_t + xsb_r*f_tb_s
                 fybp = fybp + ytb_r*f_tb_t + ysb_r*f_tb_s
                 fzbp = fzbp + ztb_r*f_tb_t + zsb_r*f_tb_s

               endif    !End of:   if(i_particleType_iip.eq.4....)then  !Floating Body Version

c               !-- Friction force between rigid objects --
c               !-- NOT ACTIVE YET IN PARALLEL CODE YET!!!! ---
c               if(1.eq.2.and.i_particleType_iip.lt.4)then   !if(1.eq.1.and.iip.lt.nbp1_local)then
c                 id_num_FB_i = i_FB_Pointer_Info(iip)
c                 pmu = friction_coeff(id_num_FB_i)
c                 !if(jjp.gt.nbfm_local)pmu = 0.0  !No friction between moving blocks   !<--------------------
c                 if(i_particleType(jjp).eq.3)pmu = 0.0  !No friction between moving blocks   !<--------------------
c                 !xtbint = znb(jjp)
c                 !ztbint = -xnb(jjp)
c                 u_parallel = (up(iip)-up(jjp))*xtb(jjp)
c     &                       +(vp(iip)-vp(jjp))*ztb(jjp)
c     &                       +(wp(iip)-wp(jjp))*ztb(jjp)
c                 f_tb = - sign(1.0,u_parallel)*pmu*facp
c                 fxbp = fxbp + f_tb*xtb(jjp)
c                 fybp = fybp + f_tb*ytb(jjp)
c                 fzbp = fzbp + f_tb*ztb(jjp)
c               end if    !End of:  if(1.eq.2.and.iip.gt.nb)then

             endif    !End of:   if(chi_X.gt.0.0.and.chi_Y.gt.0.0.and.Q_fn.gt.0.0)then
           endif    !End of:    if(i_particleType_iip.lt.4 ....)then
         else
             fxbp = 0.0
             fybp = 0.0
             fzbp = 0.0
         end if

c      !- Debugging -
c      fxbp = 0.0
c      fybp = 0.0
c      fzbp = 0.0

      end	